package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;

import java.util.Collection;
import java.util.Collections;

/**
 * A chain (a.k.a&#46; <em>linearly ordered</em>) distributive lattice.
 * 
 * <p>Elements of a chain with <var>n</var> elements are numbered from 0 to <var>n</var> &minus;
 * 1, causing an unfortunate clash between {@link #one()}, which can be obtained as
 * {@link #valueOf(String) valueOf("1")}, and the second element (starting from {@link #zero()})
 * of the chain, which can be obtained as {@link #valueOf(int) valueOf(1)}.
 * 
 */
public class Chain extends AbstractDistributiveLattice {
	private final ChainElement element[];

	private Object2ObjectOpenHashMap<String, Element> name2element = new Object2ObjectOpenHashMap<String, Element>();

	private final class ChainElement extends NamedElement {
		final int pos;

		public ChainElement( Chain chain, String name, int pos ) {
			super( chain, name );
			this.pos = pos;
		}

		public String toString() {
			if ( name != null ) return name;
			return "x" + pos;
		}

		public boolean equals( Object o ) {
			if ( o instanceof ChainElement ) {
				ChainElement chainElement = (ChainElement)o;
				return lattice == chainElement.lattice && pos == chainElement.pos;
			}
			return false;
		}

		public int hashCode() {
			return lattice.hashCode() ^ pos;
		}
	}

	/** Create a chain with the given number of elements, naming elements with the given strings.
	 * 
	 * @param n the number of elements in this chain.
	 * @param name <code>n</code>-2 names for the elements of this chain, except for
	 * {@link #zero()} and {@link #one()}.
	 */
	public Chain( int n, String... name ) {
		if ( n >= 2 ) {
			if ( name != null && name.length != n - 2 ) throw new IllegalArgumentException( "You provided " + name.length + " names for a chain of " + n + " elements" );
			element = new ChainElement[ n ];
			element[ 0 ] = new ChainElement( this, "0", 0 );
			for ( int i = 1; i < n - 1; i++ ) {
				if ( name != null && ( "0".equals( name[ i - 1 ] ) || "1".equals( name[ i - 1 ] ) ) ) throw new IllegalArgumentException( "You cannot name a chain element 0 or 1" );
				element[ i ] = new ChainElement( this, name != null ? name[ i - 1 ] : null, i );
				if ( name != null ) name2element.put( name[ i - 1 ], element[ i ] );
			}
			element[ n - 1 ] = new ChainElement( this, "1", n - 1 );
		}
		else {
			if ( name.length != 0 ) throw new IllegalArgumentException( "You provided " + name.length + " names for a chain without non trivial elements" );
			element = new ChainElement[ 1 ];
			element[ 0 ] = new ChainElement( this, "*", 0 );
			name2element.put( "*", element[ 0 ] );
		}
	}

	/**
	 * Create a chain with the given number of elements.
	 * 
	 * @param n the number of elements in this chain.
	 */
	public Chain( int n ) {
		this( n, (String[])null );
	}


	public Element valueOf( String name ) {
		Element zeroOrOne = valueOfZeroOrOne( name );
		if ( zeroOrOne != null ) return zeroOrOne;
		Element e = name2element.get( name );
		if ( e != null ) return e;
		throw new ElementNameException( name, this );
	}

	/**
	 * Return the element of this chain of given ordinal position.
	 * 
	 * <p>Note that while {@link #valueOf(int) valueOf(0)} return {@link #zero()},
	 * {@link #valueOf(int) valueOf(1)} will <strong>not</strong> return {@link #one()} unless this
	 * chain has less than three elements.
	 * 
	 * @param i the ordinal position in this chain (starting at {@link #zero()} with position 0).
	 * @return the element of this chain of ordinal position <code>i</code>.
	 */

	public Element valueOf( int i ) {
		return element[ i ];
	}

	@SuppressWarnings("unchecked")
	public Collection<Element> generators() {
		if ( element.length > 2 ) return new ObjectOpenHashSet<Element>( element, 1, element.length - 2 );
		return Collections.EMPTY_SET;
	}

	public Collection<Element> elements() {
		return new ObjectOpenHashSet<Element>( element );
	}

	public Element one() {
		return element[ element.length - 1 ];
	}

	public Element zero() {
		return element[ 0 ];
	}

	public Element join( Element... element ) {
		ensureElementsInLattice( element );

		ChainElement max = (ChainElement)zero();
		for ( Element e : element )
			if ( max.pos < ( (ChainElement)e ).pos ) max = ( (ChainElement)e );
		return max;
	}

	public Element meet( Element... element ) {
		ensureElementsInLattice( element );

		ChainElement min = (ChainElement)one();
		for ( Element e : element )
			if ( min.pos > ( (ChainElement)e ).pos ) min = ( (ChainElement)e );
		return min;
	}

	public boolean comp( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		return true;
	}

	public boolean leq( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		return ( (ChainElement)x ).pos <= ( (ChainElement)y ).pos;
	}

	public String toString() {
		return "Chain(" + element.length + ")";
	}
}
