package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/**
 * An element of a {@linkplain Lattice lattice}.
 * 
 * <p>Each lattice element has a reference to the containing lattice, and provides a number of
 * utility methods that make it easy to write down expressions and perform computations.
 * 
 */
public interface Element {
	/**
	 * Return the meet of this element and of the provided elements. In particular, upon the empty
	 * list of arguments returns this element.
	 * 
	 * @param element the elements whose meet, together with this element, has to be computed.
	 * @return the meet of the provided elements and of this element.
	 */
	Element meet( Element... element );

	/**
	 * Return the meet of this element and of the provided element.
	 * 
	 * @param element the element whose meet, together with this element, has to be computed.
	 * @return the meet of the provided element and of this element.
	 */
	Element meet( Element element );

	/**
	 * Return the join of this element and of the provided elements. In particular, upon the empty
	 * list of arguments returns this element.
	 * 
	 * @param element the elements whose join, together with this element, has to be computed.
	 * @return the join of the provided elements and of this element.
	 */
	Element join( Element... element );

	/**
	 * Return the join of this element and of the provided element.
	 * 
	 * @param element the element whose join, together with this element, has to be computed.
	 * @return the join of the provided element and of this element.
	 */
	Element join( Element element );

	/**
	 * Return whether this element is comparable with the provided element.
	 * 
	 * @param element an element.
	 * @return true if this element is {@linkplain Lattice#comp(Element, Element) comparable}
	 * with <code>element</code>, false otherwise.
	 */
	boolean comp( Element element );

	/**
	 * Return whether this element is less than or equal to the provided element in the order of
	 * {@link #lattice() the containing lattice}.
	 * 
	 * @param element an element.
	 * @return true if this element is
	 * {@linkplain Lattice#leq(Element, Element) less than or equal to} <code>element</code>,
	 * false otherwise.
	 */
	boolean leq( Element element );

	/** Return the difference of this element with another element (optional operation).
	 * 
	 * @param element an element.
	 * @return the difference between this element and <code>element</code>.
	 * @see Lattice#psdiff(Element, Element)
	 */
	Element psdiff( Element element );

	/** Return the pseudocomplement of this element relative to another element (optional operation).
	 * 
	 * @param element an element.
	 * @return the pseudocomplement of this element relative to <code>element</code>.
	 * @see Lattice#pscomp(Element, Element)
	 */
	Element pscomp( Element element );

	/** Return the symmetric difference between this element and another element (optional operation).
	 * 
	 * @param element an element.
	 * @return the symmetric difference of this element with <code>element</code>.
	 * @see Lattice#symdiff(Element, Element)
	 */
	Element symdiff( Element element );

	/**
	 * Return the lattice containing this element.
	 * 
	 * @return the lattice containing this element.
	 */
	Lattice lattice();
}
