package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;

import java.util.Collection;

/** A flat lattice: all elements are {@linkplain Lattice#comp(Element, Element) incomparable}
 * except for zero and one. Flat lattices are not distributive, unless they have less than five elements.
 * 
 * <p>Nontrivial elements of a flat lattice with <var>n</var> elements are numbered from 0 to <var>n</var> &minus;
 * 3, causing an unfortunate clash between {@link #zero()} ({@link #one()}), which can be obtained as
 * {@link #valueOf(String) valueOf("0")} ({@link #valueOf(String) valueOf("1")}), 
 * and the nontrivial element {@link #valueOf(int) valueOf(0)} ({@link #valueOf(int) valueOf(1)}).
 *
 */
public class Flat extends AbstractLattice {

	private final Object2ObjectOpenHashMap<String,Element> name2element = new Object2ObjectOpenHashMap<String,Element>(); 
	private final FlatLatticeElement zero;
	private final FlatLatticeElement one;
	private final FlatLatticeElement[] element;
	
	private final class FlatLatticeElement extends NamedElement {
		final int index;

		public FlatLatticeElement( Flat chain, String name, int pos ) {
			super( chain, name );
			this.index = pos;
		}

		public String toString() {
			if ( name != null ) return name;
			return "x" + index;
		}

		public boolean equals( Object o ) {
			if ( o instanceof FlatLatticeElement ) {
				FlatLatticeElement flatLatticeElement = (FlatLatticeElement)o;
				return lattice == flatLatticeElement.lattice && index == flatLatticeElement.index;
			}
			return false;
		}

		public int hashCode() {
			return lattice.hashCode() ^ index;
		}
	}

	/** Create a flat lattice with named elements.
	 * 
	 * @param n the number of elements in the lattice.
	 * @param name names for the lattice elements.
	 */
	public Flat( final int n, final String... name ) {
		if ( n >= 2 ) {
			if ( name != null && name.length != n - 2 ) throw new IllegalArgumentException( "You provided " + name.length + " names for a flat lattice of " + n + " elements" );
			element = new FlatLatticeElement[ n - 2 ];
			zero = new FlatLatticeElement( this, "0", -1 );
			one = new FlatLatticeElement( this, "1", Integer.MAX_VALUE );

			for ( int i = 0; i < n - 2; i++ ) {
				if ( name != null && ( "0".equals( name[ i ] ) || "1".equals( name[ i ] ) ) ) throw new IllegalArgumentException( "You cannot name a chain element 0 or 1" );
				element[ i ] = new FlatLatticeElement( this, name != null ? name[ i ] : null, i );
				if ( name != null ) name2element.put( name[ i ], element[ i ] );
			}
		}
		else {
			if ( name != null && name.length != 0 ) throw new IllegalArgumentException( "You provided " + name.length + " names for a chain without non trivial elements" );
			element = new FlatLatticeElement[ 0 ];
			zero = one = new FlatLatticeElement( this, "*", 0 );
			name2element.put( "*", zero );
		}
	}

	/** Create a flat lattice.
	 * 
	 * @param n the number of elements in the lattice.
	 */
	public Flat( int n ) {
		this( n, (String[])null );
	}
	
	@Override
	public boolean isDistributive() {
		return element.length < 3;
	}
	
	public Element valueOf( String name ) {
		Element zeroOrOne = valueOfZeroOrOne( name );
		if ( zeroOrOne != null ) return zeroOrOne;
		Element e = name2element.get( name );
		if ( e != null ) return e;
		throw new ElementNameException( name, this );
	}

	public Element valueOf( int i ) {
		return element[ i ];
	}
		
	public Collection<Element> generators() {
		return new ObjectArrayList<Element>( name2element.values() );
	}

	public Collection<Element> elements() {
		Collection<Element> result = new ObjectOpenHashSet<Element>( element );
		result.add( zero );
		result.add( one );
		return result;
	}

	public Element one() {
		return one;
	}

	public Element zero() {
		return zero;
	}

	public Element join( final Element... element ) {
		ensureElementsInLattice( element );

		Element seen = null;
		for( Element e : element ) {
			if ( e == one ) return one;
			if ( e != zero ) {
				if ( seen != null && ! e.equals( seen ) ) return one;
				seen = e;
			}
		}
		
		return seen != null ? seen : zero;
	}


	public Element meet( final Element... element ) {
		ensureElementsInLattice( element );

		Element seen = null;
		for( Element e : element ) {
			if ( e == zero ) return zero;
			if ( e != one ) {
				if ( seen != null && ! e.equals( seen ) ) return zero;
				seen = e;
			}
		}
		
		return seen != null ? seen : one;
	}

	public boolean comp( Element x, Element y ) {
		return x == zero || y == zero || x == one || y == one || x.equals( y );
	}

	public boolean leq( Element x, Element y ) {
		return x == zero || y == one || x.equals( y );
	}


}
