package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.IntArrayFIFOQueue;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.fastutil.objects.ObjectHeapSemiIndirectPriorityQueue;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;
import it.unimi.dsi.util.Interval;
import it.unimi.dsi.util.Intervals;

import java.util.Arrays;
import java.util.Collection;

import org.apache.commons.lang.ArrayUtils;

/**
 * The distributive lattice of antichains of intervals of a linear finite
 * order. It is isomorphic to the order-ideal completion of intervals ordered by
 * <em>reverse inclusion</em>, the isomorphism being given by taking the union of all principal
 * ideals generated by the elements of an antichain.
 * 
 * <p>Given an integer <var>n</var>, this class represents the lattice of interval antichains of
 * the linear order { 0, 1, &hellip;, <var>n</var> &minus; 1 }.
 * 
 * <p>This lattice plays a prominent r&ocirc;le in &ldquo;An algebra for structured text search and
 * a framework for its implementation&rdquo;, by Charles L.A. Clarke, Gordon V. Cormack, and Forbes
 * J. Burkowski, <i>Comput. J.</i>, 38(1):43&minus;56, 1995, where it is used to denote regions of text
 * satisfying a query (albeit the authors do not remark that their lattice is actually an
 * order-ideal completion). The algorithms used in this implementation are described in
 * &ldquo;Efficient lazy algorithms for minimal-interval semantics&rdquo;, by Paolo Boldi and
 * Sebastiano Vigna, <i>Proc. SPIRE 2006</i>, number 4209 in Lecture Notes in Computer Science, pages 134&minus;149. Springer&ndash;Verlag,
 * 2006, where a more mathematically oriented description of the lattice can be found.
 */

public class IntervalAntichains extends AbstractDistributiveLattice {
	/** Number of elements used to generate intervals. */
	public final int n;

	private final Antichain zero = new Antichain( this, 0, new Interval[ 0 ] );

	private final Antichain one = new Antichain( this, 1, Intervals.EMPTY_INTERVAL );
	
	/** A class representing elements of this lattice (antichains of intervals).
	 * 
	 * <p>Note that this class implements several additional operations (e.g., {@link #isContained(Element)}).
	 */
	public final static class Antichain extends AbstractElement {
		public static final boolean DEBUG = false;
		public final Interval[] interval;
		private int rank = -1;
		private int hashCode;

		public static Antichain valueOf( final IntervalAntichains lattice, final ObjectArrayList<Interval> antichain ) {
			return valueOf( lattice, antichain.size(), antichain.elements() );
		}

		public static Antichain valueOf( final IntervalAntichains lattice, final int length, final Interval... antichain ) {
			if ( length == 0 ) return lattice.zero;
			if ( length == 1 && antichain[ 0 ] == Intervals.EMPTY_INTERVAL ) return lattice.one;
			assert ! ArrayUtils.contains( antichain, Intervals.EMPTY_INTERVAL ) : antichain;
			return new Antichain( lattice, length, antichain );
		}

		public static Antichain valueOf( final IntervalAntichains lattice, final Interval... antichain ) {
			return valueOf( lattice, antichain.length, antichain );
		}

		private Antichain( IntervalAntichains lattice, final int length, final Interval... antichain ) {
			super( lattice );
			this.interval = antichain.length == length ? antichain : Arrays.copyOf( antichain, length  );
		}

		public IntervalAntichains lattice() {
			return (IntervalAntichains)lattice;
		}

		@Override
		public Antichain meet( Element... element ) {
			return (Antichain)super.meet( element );
		}

		@Override
		public Antichain meet( Element element ) {
			return meet( new Element[] { element } );
		}

		@Override
		public Antichain join( Element... element ) {
			return (Antichain)super.join( element );
		}

		@Override
		public Antichain join( Element element ) {
			return join( new Element[] { element } );
		}

		@Override
		public Antichain psdiff( Element element ) {
			return lattice().psdiff( this, element );
		}

		@Override
		public Antichain pscomp( Element element ) {
			return lattice().pscomp( this, element );
		}

		@Override
		public Antichain symdiff( Element element ) {
			return lattice().symdiff( this, element );
		}

		/** An alias for {@link #psdiff(Element)}.
		 * 
		 * @param element an element.
		 * @return the antichain formed by the intervals of this
		 * element that do not contain an interval of {@code element}.
		 */
		public Antichain doesNotContain( Element element ) {
			return psdiff( element );
		}
		
		/** Return the rank of this element.
		 * 
		 * <p>The rank will be cached. Thus, after the first call on this element
		 * this method will return the cached value.
		 * 
		 * @return the rank of this element.
		 */
		public int rank() {
			if ( this == lattice.zero() ) return 0;
			if ( this == lattice.one() ) return 1 + ((IntervalAntichains)lattice).n * ( ((IntervalAntichains)lattice).n + 1 ) / 2;
			if ( rank == -1 ) {
				final int length = interval.length;
				if( length == 0 ) rank = 0;
				else {
					int r = ( 1 + interval[ length - 1 ].left ) * lattice().n - interval[ 0 ].right * ( 1 + interval[ 0 ].left );
					for( int i = length; i-- != 1 ; ) r += interval[ i ].right * ( interval[ i - 1 ].left - interval[ i ].left );
					rank = r;
				}
			}
			return rank;
		}

		private Antichain bruteForceIsNotContained( Element element ) {
			Antichain B = (Antichain)element;
			ObjectArrayList<Interval> intervals = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );
			for ( Interval I: interval ) {
				boolean found = false;
				for ( Interval J: B.interval ) 
					if ( J.contains( I ) ) {
						found = true;
						break;
					}
				if ( ! found ) intervals.add( I );
			}
			return Antichain.valueOf( lattice(), intervals );
		}
		
		/** Return the antichain formed by the intervals of this
		 * element that are not contained in an interval of a given element.
		 * 
		 * @param element an element.
		 * @return the antichain formed by the intervals of this
		 * element that are not contained in an interval of {@code element}.
		 */
		public Antichain isNotContained( Element element ) {
			final Antichain zero = lattice().zero;
			final Antichain one = lattice().one;
			assert this.equals( zero ) == ( this == zero );
			assert this.equals( one ) == ( this == one );
			assert element.equals( zero ) == ( element == zero );
			assert element.equals( one ) == ( element == one );

			if ( element == zero ) {
				assert bruteForceIsNotContained( element ).equals( this ) : bruteForceIsNotContained( element ) + " != " + this;
				return this;
			}

			if ( this == zero || this == one ) {
				assert bruteForceIsNotContained( element ).equals( zero ) : bruteForceIsNotContained( element ) + " != " + zero;
				return zero;
			}

			if ( element == one ) {
				assert bruteForceIsNotContained( element ).equals( this ) : bruteForceIsNotContained( element ) + " != " + this;
				return this;
			}

			final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );

			final Interval[] a = interval, b = ((Antichain)element).interval;
			final int l = a.length, m = b.length;
			
			for( int i = 0, j = 0; i < l; i++ ) {
				while( j < m && b[ j ].right < a[ i ].right ) j++;
				if ( j == m ) {
					while( i < l ) result.add( a[ i++ ] );
					break;
				}
				if ( a[ i ].left < b[ j ].left ) result.add( a[ i ] );
			}


			final Antichain isNotContained = Antichain.valueOf( lattice(), result );
			assert bruteForceIsNotContained( element ).equals( isNotContained ) :  bruteForceIsNotContained( element ) + " != " + isNotContained;
			return isNotContained;
		}


		private Antichain bruteForceIsContained( Element element ) {
			Antichain B = (Antichain)element;
			ObjectArrayList<Interval> intervals = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );
			for ( Interval I: interval ) 
				for ( Interval J: B.interval ) 
					if ( J.contains( I ) ) {
						intervals.add( I );
						break;
					}
			return Antichain.valueOf( lattice(), intervals );
		}
		
		/** Return the antichain formed by the intervals of this
		 * element that are contained in some interval of a given element.
		 * 
		 * @param element an element.
		 * @return the antichain formed by the intervals of this
		 * element that are contained in some interval of {@code element}.
		 */
		public Antichain isContained( Element element ) {
			final Antichain zero = lattice().zero;
			final Antichain one = lattice().one;
			assert this.equals( zero ) == ( this == zero );
			assert this.equals( one ) == ( this == one );
			assert element.equals( zero ) == ( element == zero );
			assert element.equals( one ) == ( element == one );

			if ( this == zero || element == zero ) {
				assert bruteForceIsContained( element ).equals( zero ) : bruteForceIsContained( element ) + " != " + zero;
				return zero;
			}

			if ( this == one ) {
				assert bruteForceIsContained( element ).equals( one ) : bruteForceIsContained( element ) + " != " + one;
				return one;
			}

			if ( element == one ) {
				assert bruteForceIsContained( element ).equals( zero ) : bruteForceIsContained( element ) + " != " + zero;
				return zero;
			}

			final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );

			final Interval[] a = interval, b = ((Antichain)element).interval;
			final int l = a.length, m = b.length;
			
			for( int i = 0, j = 0; i < l; i++ ) {
				while( j < m && b[ j ].right < a[ i ].right ) j++;
				if ( j == m ) break;
				if ( b[ j ].left <= a[ i ].left ) result.add( a[ i ] );
			}

			final Antichain isContained = Antichain.valueOf( lattice(), result );
			assert bruteForceIsContained( element ).equals( isContained ) :  bruteForceIsContained( element ) + " != " + isContained;
			return isContained;
		}

		private Antichain bruteForceContains( Element element ) {
			Antichain B = (Antichain)element;
			ObjectArrayList<Interval> intervals = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );
			for ( Interval I: interval ) 
				for ( Interval J: B.interval ) 
					if ( I.contains( J ) ) {
						intervals.add( I );
						break;
					}
			return Antichain.valueOf( lattice(), intervals );
		}
		
		/** Return the antichain formed by the intervals of this
		 * element that contain an interval of a given element.
		 * 
		 * @param element an element.
		 * @return the antichain formed by the intervals of this
		 * element that contain an interval of {@code element}.
		 */
		public Antichain contains( Element element ) {
			final Antichain zero = lattice().zero;
			final Antichain one = lattice().one;
			assert this.equals( zero ) == ( this == zero );
			assert this.equals( one ) == ( this == one );
			assert element.equals( zero ) == ( element == zero );
			assert element.equals( one ) == ( element == one );

			if ( this == zero || element == zero ) {
				assert bruteForceContains( element ).equals( zero ) : bruteForceContains( element ) + " != " + zero;
				return zero;
			}

			if ( element == one ) {
				assert bruteForceContains( element ).equals( this ) : bruteForceContains( element ) + " != " + this;
				return this;
			}

			if ( this == one ) {
				assert bruteForceContains( element ).equals( zero ) : bruteForceContains( element ) + " != " + zero;
				return zero;
			}

			final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );

			final Interval[] a = interval, b = ((Antichain)element).interval;
			final int l = a.length, m = b.length;
			
			for( int i = 0, j = 0; i < l; i++ ) {
				while( j < m && b[ j ].left < a[ i ].left && b[ j ].right < a[ i ].right ) j++;
				if ( j == m ) break;
				if ( a[ i ].contains( b[ j ] ) ) result.add( a[ i ] );
			}

			final Antichain contains = Antichain.valueOf( lattice(), result );
			assert bruteForceContains( element ).equals( contains ) :  bruteForceContains( element ) + " != " + contains;
			return contains;
		}

		private Antichain bruteForceIncomparable( Element element ) {
			Antichain B = (Antichain)element;
			ObjectArrayList<Interval> intervals = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );
			for ( Interval I: interval ) {
				boolean found = false;
				for ( Interval J: B.interval ) 
					if ( J.contains( I ) || I.contains( J ) ) {
						found = true;
						break;
					}
				if ( !found ) intervals.add( I );
			}
			return Antichain.valueOf( lattice(), intervals );
		}
		
		/** Return the antichain formed by the intervals of this
		 * element that are incomparable with all intervals of a given element.
		 * 
		 * @param element an element.
		 * @return the antichain formed by the intervals of this
		 * element that are incomparable with all intervals of {@code element}.
		 */
		public Antichain incomparable( Element element ) {
			final Antichain zero = lattice().zero;
			final Antichain one = lattice().one;
			assert this.equals( zero ) == ( this == zero );
			assert this.equals( one ) == ( this == one );
			assert element.equals( zero ) == ( element == zero );
			assert element.equals( one ) == ( element == one );

			if ( this == zero || element == one ) {
				assert bruteForceIncomparable( element ).equals( zero ) : bruteForceIncomparable( element ) + " != " + zero;
				return zero;
			}

			if ( element == zero ) {
				assert bruteForceIncomparable( element ).equals( this ) : bruteForceIncomparable( element ) + " != " + this;
				return this;
			}

			if ( this == one ) {
				assert bruteForceIncomparable( element ).equals( zero ) : bruteForceIncomparable( element ) + " != " + zero;
				return zero;
			}

			final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );

			final Interval[] a = interval, b = ((Antichain)element).interval;
			final int l = a.length, m = b.length;
			
			main: for( int i = 0, j = 0; i < l; i++ ) {
				while( b[ j ].left < a[ i ].left ) { 
					while ( b[ j ].right >= a[ i ].right ) if ( ++i == l ) break main;
					if ( ++j == m ) {
						while( i < l ) result.add( a[ i++ ] );
						break main;
					}
				}

				if ( b[ j ].left > a[ i ].left && b[ j ].right > a[ i ].right ) result.add ( a[ i ] );
			}

			final Antichain incomparable = Antichain.valueOf( lattice(), result );
			assert bruteForceIncomparable( element ).equals( incomparable ) :  bruteForceIncomparable( element ) + " != " + incomparable;
			return incomparable;
		}

		private Antichain bruteForceSetIntersection( Element element ) {
			Antichain B = (Antichain)element;
			ObjectOpenHashSet<Interval> intervalSet = new ObjectOpenHashSet<Interval>( interval );
			intervalSet.retainAll( new ObjectOpenHashSet<Interval>( B.interval ) );
			Interval[] antichain = intervalSet.toArray( new Interval[ 0 ] );
			Arrays.sort( antichain, Intervals.ENDS_BEFORE );
			return Antichain.valueOf( lattice(), antichain );
		}

		/** Return the set intersection of this element with another element.
		 * 
		 * @param element an element.
		 * @return the set intersection of this element with {@code element}.
		 */
		public Antichain setIntersection( Element element ) {
			final Antichain zero = lattice().zero;
			final Antichain one = lattice().one;
			assert this.equals( zero ) == ( this == zero );
			assert this.equals( one ) == ( this == one );
			assert element.equals( zero ) == ( element == zero );
			assert element.equals( one ) == ( element == one );

			if ( this.equals( element ) ) return this;

			if ( this == zero || element == zero || element == one || this == one ) {
				assert bruteForceSetIntersection( element ).equals( zero ) : bruteForceSetIntersection( element ) + " != " + zero;
				return zero;
			}

			final Interval[] a = interval, b = ((Antichain)element).interval;
			final int l = a.length, m = b.length;
			final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ lattice().n ], 0 );
			
			for( int i = 0, j = 0; i < l && j < m; ) {
				if ( a[ i ].left == b[ j ].left ) {
					if ( a[ i ].right == b[ j ].right ) result.add( a[ i ] );
					i++; 
					j++;
				}
				else if ( a[ i ].left < b[ j ].left ) i++;
				else j++;
			}

			Antichain intersection = Antichain.valueOf( lattice(), result );
			assert bruteForceSetIntersection( element ).equals( intersection ) : bruteForceSetIntersection( element ) + " != " + intersection;
			return intersection;
		}

		@Override
		public boolean equals( final Object o ) {
			if ( ! ( o instanceof Antichain ) ) return false;
			final Antichain antichain = (Antichain)o;
			return antichain.lattice == lattice && Arrays.equals( antichain.interval, interval );
		}

		@Override
		public int hashCode() {
			if ( hashCode == 0 ) hashCode = lattice.hashCode() ^ Arrays.hashCode( interval );
			return hashCode;
		}		
		
		@Override
		public String toString() {
			if ( interval.length == 0 ) return "\u2205";
			StringBuilder s = new StringBuilder( "{" );
			for ( int i = 0; i < interval.length; i++ ) {
				if ( i != 0 ) s.append( ',' );
				s.append( interval[ i ] );
			}
			return s.append( '}' ).toString();
		}
	}

	public IntervalAntichains( final int n ) {
		this.n = n;
	}

	@Override
	public Antichain meet( final Element... element ) {
		if ( element.length == 0 ) return one;
		if ( element.length == 1 ) return (Antichain)element[ 0 ];

		final Interval[] curr = new Interval[ element.length ];
		final ObjectHeapSemiIndirectPriorityQueue<Interval> queue = new ObjectHeapSemiIndirectPriorityQueue<Interval>( curr, Intervals.STARTS_BEFORE_OR_PROLONGS );

		final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ n ], 0 );
		final int[] pos = new int[ element.length ];
		int nextLeft, nextRight, maxRight = -1;

		final Interval[][] antichain = new Interval[ element.length ][];

		for ( int i = 0; i < element.length; i++ ) {
			if ( element[ i ] == zero ) return zero; // zero in a meet;
			if ( element[ i ] != one ) {
				antichain[ i ] = ((Antichain)element[ i ]).interval;
				curr[ i ] = antichain[ i ][ 0 ];

				queue.enqueue( i );
				maxRight = Math.max( maxRight, curr[ i ].right );
				pos[ i ]++;
			}
		}

		if ( queue.isEmpty() ) return one; // All elements are ones

		int first;

		nextLeft = -1;
		nextRight = -1;

		for ( ;; ) {

			while ( ( curr[ first = queue.first() ].left ) == nextLeft ) {
				if ( pos[ first ] == antichain[ first ].length )
					return Antichain.valueOf( this, result );

				curr[ first ] = antichain[ first ][ pos[ first ]++ ];
				maxRight = Math.max( maxRight, curr[ first ].right );
				queue.changed();
			}

			do {
				first = queue.first();
				nextLeft = curr[ first ].left;
				nextRight = maxRight;

				if ( curr[ first ].right == maxRight ) break; // If the top interval is the whole output, we can emit it
					
				if ( pos[ first ] == antichain[ first ].length ) {
					result.add( Interval.valueOf( nextLeft, nextRight ) );
					return Antichain.valueOf( this, result );
				}

				curr[ first ] = antichain[ first ][ pos[ first ]++ ];
				maxRight = Math.max( maxRight, curr[ first ].right );
				queue.changed();
			} while ( maxRight == nextRight );

			result.add( Interval.valueOf( nextLeft, nextRight ) );
		}
	}

	@Override
	public Antichain join( final Element... element ) {
		if ( element.length == 0 ) return zero;
		if ( element.length == 1 ) return (Antichain)element[ 0 ];

		final Interval[] curr = new Interval[ element.length ];
		final ObjectHeapSemiIndirectPriorityQueue<Interval> queue = new ObjectHeapSemiIndirectPriorityQueue<Interval>( curr, Intervals.ENDS_BEFORE_OR_IS_SUFFIX );
		final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ n ], 0 );
		final int[] pos = new int[ element.length ];
		final Interval[][] antichain = new Interval[ element.length ][];

		for ( int i = 0; i < element.length; i++ ) {
			if ( element[ i ] == one ) return one; // one in a join;
			if ( element[ i ] != zero ) {
				antichain[ i ] = ((Antichain)element[ i ]).interval;
				curr[ i ] = antichain[ i ][ 0 ];

				queue.enqueue( i );
				pos[ i ]++;
			}
		}

		if ( queue.isEmpty() ) return zero; // All elements are zeroes

		int first;
		Interval top;

		while( ! queue.isEmpty() ) {
			result.add( top = curr[ queue.first() ] );
			
			do {
				first = queue.first();
				if ( pos[ first ] < antichain[ first ].length ) {
					curr[ first ] = antichain[ first ][ pos[ first ]++ ];
					queue.changed();
				}
				else {
					queue.dequeue();
					if ( queue.isEmpty() ) break;
				}
			} while ( curr[ queue.first() ].left <= top.left );
		}
		
		return Antichain.valueOf( this, result );

	}

	@Override
	public Collection<Element> generators() {
		final ObjectArrayList<Element> generators = ObjectArrayList.wrap( new Element[ n ], 0 );
		for ( int i = 0; i < n; i++ ) generators.add( Antichain.valueOf( this, Interval.valueOf( i ) ) );
		return generators;
	}

	/**
	 * Parse an antichain, returning the corresponding element.
	 * 
	 * <p>An <em>interval</em> is described (using Hoare's notation) by a pair of integers,
	 * separated by two dots and surrounded by square brackets. Thus, <code>[1..3]</code> denotes
	 * the interval containing the integers 1, 2, and 3. The shortcut <code>[<var>x</var>]</code>
	 * can be used to denote the singleton interval containing <var>x</var>.
	 * 
	 * <p>An <em>interval antichain</em> is described by a list of intervals separated by commas.
	 * Of course, no interval in the list must contain another interval. Intervals can be in any
	 * order. For instance, <code>[1..3], [2..4], [0]</code> is a legal antichain.
	 * 
	 * <p>Spaces are not significative, but the two dots in an interval must be consecutive.
	 * 
	 * @param name the description of an antichain.
	 * @return the corresponding element of this lattice.
	 */

	@Override
	public Antichain valueOf( String name ) {
		String[] interval = name.split( "," );
		ObjectArrayList<Interval> antichain = ObjectArrayList.wrap( new Interval[ n ], 0 );
		for ( String t : interval ) {
			String i = t.trim();
			if ( i.charAt( 0 ) != '[' || i.charAt( i.length() - 1 ) != ']' ) throw new IllegalArgumentException();
			final String[] extreme = i.substring( 1, i.length() - 1 ).split( "\\.\\." );
			final Interval newInterval;
			if ( extreme.length == 1 ) newInterval = Interval.valueOf( Integer.parseInt( extreme[ 0 ] ) ); // singleton
			else newInterval = Interval.valueOf( Integer.parseInt( extreme[ 0 ] ), Integer.parseInt( extreme[ 1 ] ) );
			if ( newInterval.right < 0 || newInterval.left >= n ) throw new IllegalArgumentException( "Interval out of range: " + newInterval ); 

			for ( Interval oldInterval : antichain )
				if ( oldInterval.contains( newInterval ) || newInterval.contains( oldInterval ) ) throw new IllegalArgumentException( "Intervals " + oldInterval + " and " + newInterval
						+ " are comparable" );

			antichain.add( newInterval );
		}
		Arrays.sort( antichain.elements(), 0, antichain.size(), Intervals.ENDS_BEFORE );
		return Antichain.valueOf( this, antichain );
	}

	/** Return a singleton representing the zero of this lattice (the empty antichain).
	 * 
	 * <p>The element returned by this method is a singleton: it has a single
	 * instance. All lattice scomputation methods will return this object when they
	 * have to return zero. As a consequence, it is possible to test for zero
	 * using {@code ==} instead of {@code equals()}.
	 *
	 * @return a singleton representing the zero of this lattice (the empty antichain).
	 */
	@Override
	public Antichain zero() {
		return zero;
	}

	/** Return a singleton representing the zero of this lattice (the antichain formed by the empty interval).
	 *
	 * <p>The element returned by this method is a singleton: it has a single
	 * instance. All lattice computation methods will return this object when they
	 * have to return one. As a consequence, it is possible to test for one
	 * using {@code ==} instead of {@code equals()}.
	 *
	 * @return a singleton representing the zero of this lattice (the antichain formed by the empty interval).
	 */
	@Override
	public Antichain one() {
		return one;
	}
	
	/**
	 * Return whether the two provided elements are comparable.
	 * 
	 * @return true if x &le; y or y &le; x, false otherwise.
	 */
	@Override
	public boolean comp( Element x, Element y ) {
		return x.leq( y ) || y.leq( x );
	}

	/**
	 * Return whether an element is less than or equal to another element in the natural order of
	 * this lattice.
	 * 
	 * <p>This implementation is based on a simple linear greedy algorithm that uses an explicit
	 * definition (an antichain <var>A</var> is smaller than or equal to an antichain <var>B</var>
	 * iff for every interval in <var>A</var> there is a corresponding smaller interval in <var>B</var>).
	 * 
	 * @return true iff <code>x</code> &le; <code>y</code>.
	 */
	
	@Override
	public boolean leq( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		assert x.equals( zero ) == ( x == zero );
		assert x.equals( one ) == ( x == one );
		assert y.equals( zero ) == ( y == zero );
		assert y.equals( one ) == ( y == one );

		if ( y == one || x == zero ) return true;
		final Antichain ax = ((Antichain)x), bx = ((Antichain)y);
		final Interval[] a = ax.interval, b = bx.interval;
		
		for( int i = a.length, j = b.length - 1; i-- != 0; ) {
			while( j != -1 && b[ j ].right > a[ i ].right ) j--;
			if ( j == -1 || b[ j ].left < a[ i ].left ) return false;
		}
		return true;
	}

	/** Return an antichain equal to the first element from which all intervals
	 * containing some interval of the second element have been eliminated.
	 * 
	 * <p>In the case of this lattice, the difference assumes the simple form above, and it
	 * can be easily computed with a linear greedy algorithm.
	 * 
	 * @return an antichain equal to <code>x</code> with intervals
	 * containing some interval of <code>y</code> removed.
	 */
	@Override
	public Antichain psdiff( final Element x, final Element y ) {
		ensureElementsInLattice( x, y );
		assert x.equals( zero ) == ( x == zero );
		assert x.equals( one ) == ( x == one );
		assert y.equals( zero ) == ( y == zero );
		assert y.equals( one ) == ( y == one );

		if ( x == zero || y == one ) {
			assert super.psdiff( x, y ).equals( zero ) : super.psdiff( x, y ) + " != " + zero;
			return zero;
		}

		if ( y == zero ) {
			assert super.psdiff( x, y ).equals( x ) : super.psdiff( x, y ) + " != " + x;
			return (Antichain)x;
		}

		if ( x == one ) {
			assert super.psdiff( x, y ).equals( one ) : super.psdiff( x, y ) + " != " + one;
			return one;
		}

		final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ n ], 0 );

		final Interval[] a = ((Antichain)x).interval, b = ((Antichain)y).interval;
		final int l = a.length, m = b.length;
		
		for( int i = 0, j = 0; i < l; i++ ) {
			while( j < m && b[ j ].left < a[ i ].left ) j++;
			if ( j == m ) {
				while( i < l ) result.add( a[ i++ ] );
				break;
			}
			if ( a[ i ].right < b[ j ].right ) result.add( a[ i ] );
		}


		final Antichain psdiff = Antichain.valueOf( this, result );
		assert super.psdiff( x, y ).equals( psdiff ) :  super.psdiff( x, y ) + " != " + psdiff;
		return psdiff;
	}
	
	/** Add either the singleton antichain consisting of the interval [left..right] (if left&geq;right), or
	 *  the antichain of singleton intervals from right (inclusive) to left (inclusive).
	 *  
	 * @param left one extreme.
	 * @param right the other extreme.
	 * @param result the list to which to add intervals.
	 */
	private void addFromUpTo( int left, int right, ObjectArrayList<Interval> result ) {
		if ( left <= right ) result.add( Interval.valueOf( left, right ) );
		else for ( int x = right; x <= left; x++ ) result.add( Interval.valueOf( x ) );
	}
	
	/** Add the antichain of all singleton intervals from the given integer (inclusive) up to n-1.
	 * 
	 * @param left the left extreme.
	 * @param result the list to which to add intervals.
	 */
	private void addFrom( int left, ObjectArrayList<Interval> result ) {
		for ( int x = left; x <= n - 1; x++ ) result.add( Interval.valueOf( x ) );
	}
	
	/** Add the antichain of all singleton intervals from 0 up to the given integer (inclusive).
	 * 
	 * @param right the right extreme.
	 * @param result the list to which to add intervals.
	 */
	private void addUpto( int right, ObjectArrayList<Interval> result ) {
		for ( int x = 0; x <= right; x++ ) result.add( Interval.valueOf( x ) );
	}

	
	@Override
	public Antichain pscomp( Element x, Element y ) {
		assert x.equals( zero ) == ( x == zero );
		assert x.equals( one ) == ( x == one );
		assert y.equals( zero ) == ( y == zero );
		assert y.equals( one ) == ( y == one );

		if ( x == zero || y == one ) {
			assert super.pscomp( x, y ).equals( one ) : super.pscomp( x, y ) + " != " + one;
			return one;
		}

		if ( y == zero ) {
			assert super.pscomp( x, y ).equals( zero ) : super.pscomp( x, y ) + " != " + zero;
			return zero;
		}

		if ( x == one ) {
			assert super.pscomp( x, y ).equals( y ) : super.pscomp( x, y ) + " != " + y;
			return (Antichain)y;
		}

		final Antichain A = (Antichain)x;
		final Antichain B = (Antichain)y;
		
		final int k = B.interval.length;

		// Compute hat T
		final IntArrayFIFOQueue hatT = new IntArrayFIFOQueue();

		for ( int i = 1, j = 0; i < k; i++ ) {
			// Skip empty intervals
			if ( B.interval[ i - 1 ].left + 1 <= B.interval[ i ].right - 1 ) {
				// Look for contained interval in A
				while( j < A.interval.length && A.interval[ j ].left < B.interval[ i - 1 ].left + 1 ) j++;
				if ( j == A.interval.length ) break;
				// Check containement
				if ( A.interval[ j ].right <= B.interval[ i ].right - 1 ) hatT.enqueue( i );
			}
		}
		
		final ObjectArrayList<Interval> result = ObjectArrayList.wrap( new Interval[ n ], 0 );
		
		final boolean L = ( B.interval[ k - 1 ].left + 1 <= n - 1 ? Antichain.valueOf( this, Interval.valueOf( B.interval[ k - 1 ].left + 1, n - 1 ) ) : one() ).leq( A );   
		final boolean R = ( 0 <= B.interval[ 0 ].right - 1 ? Antichain.valueOf( this, Interval.valueOf( 0, B.interval[ 0 ].right - 1 ) ) : one() ).leq( A );  

		if ( hatT.isEmpty() ) {
			if ( L & R ) addFromUpTo(  B.interval[ k - 1 ].left, B.interval[ 0 ].right, result ); 
			else if ( L ) addUpto( B.interval[ k - 1 ].left, result );
			else if ( R ) addFrom( B.interval[ 0 ].right, result );
			else return one();
		}
		else {
			final int iHatMin = hatT.firstInt();
			final int iHatMax = hatT.lastInt();

			// Compute U^-
			if ( R ) addFromUpTo( B.interval[ iHatMin - 1 ].left, B.interval[ 0 ].right, result );
			else addUpto( B.interval[ iHatMin - 1 ].left, result );

			// Compute central part
			int pred = hatT.dequeueInt();
			while( ! hatT.isEmpty() ) {
				final int i = hatT.dequeueInt();
				addFromUpTo( B.interval[ i - 1 ].left, B.interval[ pred ].right, result );
				pred = i;
			}

			// Compute U^+
			if ( L ) addFromUpTo( B.interval[ k - 1 ].left, B.interval[ iHatMax ].right, result ); 
			else addFrom( B.interval[ iHatMax ].right, result );
		}

		final Antichain pscomp = Antichain.valueOf( this, result );

		assert super.pscomp( A, B ).equals( pscomp ) : super.pscomp( A, B ) + " != " + pscomp;
		return pscomp;
	}

	@Override
	public Antichain symdiff( final Element x, final Element y ) {
		return psdiff( x, y ).join( psdiff( y, x ) );
	}

	private void enumerate( ObjectArrayList<Interval> antichain, ObjectArrayList<Element> result, final int start, final int end ) {
		result.add( Antichain.valueOf( this, antichain.toArray( new Interval[ antichain.size() ] ) ) );
		for( int l = start; l < n; l++ )
			for( int r = Math.max( end, l ); r < n; r++ ) {
				antichain.add( Interval.valueOf( l, r ) );
				enumerate( antichain, result, l + 1, r + 1 );
				antichain.remove( antichain.size() - 1 );
			}
	}
	
	/** Generate iteratively all elements of this lattice.
	 * 
	 * <p>This methods uses the CAT enumeration algorithm for interval antichains by Boldi and Vigna.
	 * 
	 * @return the set of elements of this lattice. 
	 */
	@Override
	public Collection<Element> elements() {
		ObjectArrayList<Element> result = new ObjectArrayList<Element>();
		enumerate( ObjectArrayList.wrap( new Interval[ n ], 0 ), result, 0, 0 );
		result.add( one );
		assert new ObjectOpenHashSet<Element>( super.elements() ).equals( new ObjectOpenHashSet<Element>( result ) ) : super.elements() + " != " + result;
		return result;
	}
}
