package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import java.util.Collection;
import java.util.Map;
import java.util.Set;

/** A lattice with zero and one (sometimes called a <em>bounded lattice</em>). This
 * is equivalent to a partial order with all finite sups and infs.
 * 
 * <p>The method {@link #isDistributive()} can be used to discover whether distributivity holds.  
 */

public interface Lattice {
	/** Use UTF-8 symbols for operators in all outputs. This constant is
	 * settable using the Boolean system property <code>it.unimi.dsi.lama4j.utf8</code>. */
	public final static boolean UTF8 = java.lang.Boolean.getBoolean( "it.unimi.dsi.lama4j.utf8" );
	/** Use ring notation (<code>+</code> and <code>*</code> instead of <code>|</code> and <code>&amp;</code>)
	 * in all output. This constant is settable using the Boolean system property <code>it.unimi.dsi.lama4j.ring</code>. */
	public final static boolean RING = java.lang.Boolean.getBoolean( "it.unimi.dsi.lama4j.ring" );
	
	/** Return true if this lattice is distributive. 
	 * 
	 * @return true if this lattice is distributive.
	 */
	public boolean isDistributive();
	
	/**
	 * Return the meet of the provided elements. In particular, upon the empty list of arguments
	 * returns {@link #one() one}, and upon a singleton list the only specified element.
	 * 
	 * @param element the elements whose meet has to be computed.
	 * @return the meet of the provided elements.
	 */
	public Element meet( Element... element );

	/**
	 * Return the join of the provided elements. In particular, upon the empty list of arguments
	 * returns {@link #zero() zero}, and upon a singleton list the only specified element.
	 * 
	 * @param element the elements whose join has to be computed.
	 * @return the join of the provided elements.
	 */
	public Element join( Element... element );

	/**
	 * Return a collection of generators for the lattice. The set will not include {@link #zero() zero} or
	 * {@link #one() one}. There is no guarantee of freeness or minimality.
	 * 
	 * @return a collection of generators.
	 */
	public Collection<Element> generators();

	/**
	 * Return all elements of this lattice (optional operation).
	 * 
	 * <p>This operation might not implemented, for instance, in infinite lattices.
	 * 
	 * @return the collection of all elements of this lattice.
	 */
	public Collection<Element> elements();

	/**
	 * Return an element of this lattice, given its name.
	 * 
	 * <p>Certain lattices make it possible to define names for elements. This method returns the
	 * element corresponding to the provided name.
	 * 
	 * @param name the name of an element of this lattice.
	 * @return the element of this lattice named <code>name</code>.
	 * @throws ElementNameException if the provided name does not match any
	 * element of this lattice.
	 */
	public Element valueOf( String name );

	/**
	 * Return the zero of this lattice.
	 *
	 * <p>Note that there is no guarantee that the returned element is the
	 * only element representing zero in this lattice. Other zeroes
	 * may arise from computations, but they will always be {@linkplain Object#equals(java.lang.Object) equal}
	 * to the element returned by this method.
	 * 
	 * @return the zero of this lattice.
	 */
	public Element zero();

	/**
	 * Return the one of this lattice.
	 * 
	 * <p>Note that there is no guarantee that the returned element is the
	 * only element representing one in this lattice. Other ones
	 * may arise from computations, but they will always be {@linkplain Object#equals(java.lang.Object) equal}
	 * to the element returned by this method.
	 * 
	 * @return the one of this lattice.
	 */
	public Element one();

	/**
	 * Return whether the two provided elements are comparable.
	 * 
	 * @param x an element.
	 * @param y another element.
	 * @return true if x &le; y or y &le; x, false otherwise.
	 */
	public boolean comp( Element x, Element y );

	/**
	 * Return whether an element is less than or equal to another element in the natural order of
	 * this lattice.
	 * 
	 * @param x an element.
	 * @param y another element.
	 * @return true if x &le; y, false otherwise.
	 */
	public boolean leq( Element x, Element y );

	/** Return the Brouwerian pseudo-difference of two elements (optional operation). 
	 * 
	 * <p>The <em>(Brouwerian) pseudo-difference</em> of <var>x</var> and <var>y</var>, usually denoted by
	 * <var>x</var> &minus; <var>y</var>, is defined by 
	 * a Galois connection with the join operation (categorically speaking,
	 * an adjunction):
	 * <div style="text-align: center; margin: 0 1em">
	 * <var>x</var> &minus; <var>y</var> &le; <var>t</var> &nbsp;iff&nbsp; <var>x</var> &le; <var>y</var> &#x2228; <var>t</var> &nbsp;for all <var>t</var>.
	 * </div>
	 * <p>If the Galois connection exists, this lattice is endowed with the structure of a <em>Brouwerian algebra</em>.
	 * In that case, if the lattice is finite
	 * <div style="text-align: center; margin: 0 1em">
	 * <var>x</var> &minus; <var>y</var> = inf { <var>z</var> | <var>x</var> &le; <var>y</var> &#x2228; <var>z</var> },
	 * </div>
	 * <p>and the lattice is necessarily distributive. Conversely, all 
	 * finite distributive lattices are Brouwerian algebras, with <var>x</var> &minus; <var>y</var> defined as above. 
	 * 
	 * @param x an element.
	 * @param y another element.
	 * @return <code>x</code> &minus; <code>y</code>.
	 * @throws UnsupportedOperationException if this lattice is not Browerian.
	 */
	public Element psdiff( Element x, Element y );

	/** Return the pseudocomplement of the first element relative to the second element (optional operation).
	 *
	 * <p>The <em>pseudocomplement</em> of <var>x</var> relative to <var>y</var>, denoted by
	 * <var>x</var> &#8658; <var>y</var>, is defined by a Galois connection with the meet operation (categorically speaking,
	 * an adjunction):
	 * <div style="text-align: center; margin: 0 1em">
	 * <var>t</var> &le; <var>y</var> &#8658; <var>x</var> &nbsp;iff&nbsp; <var>x</var> &#x2227; <var>t</var> &le; <var>y</var> &nbsp;for all <var>t</var>.
	 * </div>
	 * 
	 * <p>If the Galois connection exists, this lattice is endowed with the structure of a <em>Heyting algebra</em>.
	 * In that case, if the lattice is finite
	 * <div style="text-align: center; margin: 0 1em">
	 * <var>x</var> &#8658; <var>y</var> = sup { <var>z</var> | <var>x</var> &#x2227; <var>z</var> &le; <var>y</var> },
	 * </div>
	 * <p>and the lattice is necessarily distributive. Conversely, all 
	 * finite distributive lattices are Heyting algebras.
	 * 
	 * @param x an element.
	 * @param y another element.
	 * @return <code>x</code> &#8658; <code>y</code>.
	 */
	public Element pscomp( Element x, Element y );

	/** Return the symmetric difference of the arguments, that is,
	 * {@code psdiff(x,y).join(psdiff(y,x))}.
	 * 
	 * @see Element#join(Element)
	 * @see Lattice#psdiff(Element, Element)
	 * @param x an element.
	 * @param y another element.
	 * @return <code>x</code> &#916; <code>y</code>.
	 */
	
	public Element symdiff( final Element x, final Element y );	
	
	/** Return the covering relation of this lattice.
	 *
	 * <p>The <em>covering relation</em> of a lattice relates elements <var>x</var>, <var>y</var>
	 * such that there is no element strictly between <var>x</var> and <var>y</var>. In can be interpreted
	 * as a graph and {@linkplain Lattices#generateDot(Lattice, Map, java.io.PrintWriter, boolean) drawn}, 
	 * resulting in the <em>Hasse diagram</em> of the lattice.
	 * 
	 * @return a map from elements to set of elements representing the covering relation.
	 */
	public Map<Element,Set<Element>> coveringRelation();

}
