package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

/** The opposite (a.k.a. <em>dual</em>) of a given lattice. 
 * 
 * <p>Instances of this class wrap a given lattice, dualising all 
 * its operations.
 *    
 * <p>Since the lattice axioms (including distributivity) are self-dual, 
 * by reversing the order relation
 * (or, equivalently, exchanging meets and joins) we get a new lattice. 
 * Note that the definitions of {@linkplain Lattice#psdiff(Element, Element) difference} and
 * {@linkplain Lattice#pscomp(Element, Element) relative pseudocomplement} are dual, 
 * so a lattice is a Browerian algebra iff its opposite is a Heyting algebra.
 */
public class Op implements Lattice {
	/** The dualised lattice.. */
	private final Lattice lattice;
	private final Element zero;
	private final Element one ;
	
	private static final class DualisedElement extends AbstractElement {
		final public Element element;

		protected DualisedElement( Lattice lattice, Element element ) {
			super( lattice );
			this.element = element;
		}
		
		@Override
		public String toString() {
			if ( element.equals( element.lattice().zero() ) ) return "1";
			if ( element.equals( element.lattice().one() ) ) return "0";
			return element.toString();
		}
		
		@Override
		public int hashCode() {
			return element.hashCode();
		}
		
		@Override
		public boolean equals( final Object o ) {
			return o instanceof DualisedElement && ((DualisedElement)o).element.equals( element );
		}
	}
	
	public Op( final Lattice lattice ) {
		this.lattice = lattice;
		this.zero = new DualisedElement( this, lattice.one() );
		this.one = new DualisedElement( this, lattice.zero() );
	}

	public boolean isDistributive() {
		return lattice.isDistributive();
	}

	public Element meet( final Element... element ) {
		for( int i = element.length; i-- != 0; ) element[ i ] = ((DualisedElement)element[ i ]).element;
		return new DualisedElement( this, lattice.join( element ) );
	}

	public Element join( final Element... element ) {
		for( int i = element.length; i-- != 0; ) element[ i ] = ((DualisedElement)element[ i ]).element;
		return new DualisedElement( this, lattice.meet( element ) );
	}

	public Collection<Element> generators() {
		final ObjectArrayList<Element> generators = ObjectArrayList.wrap( new Element[ lattice.generators().size() ], 0 ); 
		for( Element e: lattice.generators() ) generators.add( new DualisedElement( this, e ) );
		return generators;
	}

	public Collection<Element> elements() {
		final ObjectArrayList<Element> elements = new ObjectArrayList<Element>( lattice.elements().size() ); 
		for( Element e: lattice.elements() ) elements.add( new DualisedElement( this, e ) );
		return elements;
	}

	/** Return a dualised element by name.
	 * 
	 * <p>Dual elements have the same name of the corresponding 
	 * elements of the dualised lattice,
	 * except for zero and one, which have the standard names.
	 */
	
	public Element valueOf( final String name ) {
		return new DualisedElement( this, lattice.valueOf( name ) );
	}

	public Element zero() {
		return zero;
	}

	public Element one() {
		return one;
	}

	public boolean comp( final Element x, final Element y ) {
		return lattice.comp( ((DualisedElement)x).element, ((DualisedElement)y).element );
	}

	public boolean leq( final Element x, final Element y ) {
		return lattice.leq( ((DualisedElement)y).element, ((DualisedElement)x).element );
	}

	public Element psdiff( final Element x, final Element y ) {
		return new DualisedElement( this, lattice.pscomp( ((DualisedElement)y).element, ((DualisedElement)x).element ) );
	}

	public Element pscomp( final Element x, final Element y ) {
		return new DualisedElement( this, lattice.psdiff( ((DualisedElement)y).element, ((DualisedElement)x).element ) );
	}

	public Element symdiff( final Element x, final Element y ) {
		return new DualisedElement( this, lattice.symdiff( ((DualisedElement)x).element, ((DualisedElement)y).element ) );
	}

	/** Compute the covering relation by transposing the covering relation of the
	 * dualised lattice.
	 * 
	 */
	
	public Map<Element, Set<Element>> coveringRelation() {
		Map<Element, Set<Element>> covering = lattice.coveringRelation();
		Map<Element, Set<Element>> dualCovering = new Object2ObjectOpenHashMap<Element, Set<Element>>();
		for( Map.Entry<Element, Set<Element>> e: covering.entrySet() ) {
			final Element eOp = new DualisedElement( this, e.getKey() );
			for( Element y : e.getValue() ) {
				final Element yOp = new DualisedElement( this, y );
				Set<Element> set = dualCovering.get( yOp );
				if ( set == null ) dualCovering.put( yOp, set = new ObjectOpenHashSet<Element>() );
				set.add( eOp );
			}
		}
		
		return dualCovering;
	}
}
