package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;

import java.util.Collection;
import java.util.Set;

import org.junit.Assert;
import org.junit.Ignore;

/** Static utility methods for lattice tests. */

@Ignore
public class LaMa4JTestCase {

	/** Assert that the provided lattice is actually 
	 * (possibly {@linkplain Lattice#isDistributive() distributive})
	 * lattice by testing all equations that must be true in all lattices. Additionally,
	 * interdipendencies between order-related and algebraic methods are verified, and difference
	 * and pseudocomplement are tested against the standard implementation (i.e., that of {@link it.unimi.dsi.lama4j.AbstractLattice}).
	 * @param lattice a lattice.
	 */
	
	@SuppressWarnings("boxing")
	public static void assertLattice( final Lattice lattice ) {
		Collection<Element> elements = lattice.elements();
		Element zero = lattice.zero();
		Element one = lattice.one();
		
		// Verify zero/one
		for( Element e : elements ) {
			Assert.assertEquals( zero, e.meet( zero ) );
			Assert.assertEquals( e, e.join( zero ) );
			Assert.assertEquals( one, e.join( one ) );
			Assert.assertEquals( e, e.meet( one ) );
		}
		
		// Verify commutativity
		for( Element x : elements ) 
			for( Element y : elements ) {
				Assert.assertEquals( x.meet( y ), y.meet( x ) );
				Assert.assertEquals( x.join( y ), y.join( x ) );
			}
				
		// Verify associativity
		for( Element x : elements ) 
			for( Element y : elements ) 
				for( Element z : elements ) { 
					Assert.assertEquals( lattice.meet( x, y, z ), x.meet( y ).meet( z ) );
					Assert.assertEquals( lattice.meet( x, y, z ), x.meet( y.meet( z ) ) );

					Assert.assertEquals( lattice.join( x, y, z ), x.join( y ).join( z ) );
					Assert.assertEquals( lattice.join( x, y, z ), x.join( y.join( z ) ) );
				}

		// Verify idempotence
		for( Element x : elements ) {
			Assert.assertEquals( x, x.meet( x ) );
			Assert.assertEquals( x, x.join( x ) );
		}
				
		// Verify distributivity
		if ( lattice.isDistributive() ) {
			for( Element x : elements ) 
				for( Element y : elements ) 
					for( Element z : elements ) { 
						Assert.assertEquals( x.meet( z ).join( y.meet( z ) ), x.join( y ).meet( z ) );
						Assert.assertEquals( x.join( z ).meet( y.join( z ) ), x.meet( y ).join( z ) );
					}
		}

		// Verify comp and leq
		for( Element x : elements ) 
			for( Element y : elements ) {
				Assert.assertEquals( x + ", " + y, x.leq( y ), x.meet( y ).equals( x )  );
				Assert.assertEquals( x + ", " + y, x.leq( y ), x.join( y ).equals( y )  );
				Assert.assertEquals( x + ", " + y, x.comp( y ), x.leq( y ) || y.leq( x ) );
			}
		
		try {
			// Verify psdiff
			for( Element x : elements ) 
				for( Element y : elements ) {
					Element curr = lattice.one();
					for( Element e: lattice.elements() )
						if ( x.leq( y.join( e ) ) ) curr = curr.meet(  e );
					Assert.assertEquals( x + " - " + y, curr, x.psdiff( y ) );
				}
			
			// Verify pseudocomplements
			for( Element x : elements ) 
				for( Element y : elements ) {
					Element curr = lattice.zero();
					for( Element e: lattice.elements() )
						if ( x.meet( e ).leq( y ) ) curr = curr.join(  e );
					Assert.assertEquals( x + " => " + y, curr, x.pscomp( y ) );
				}
		}
		catch( UnsupportedOperationException ok ) {}
		
		// Verify the covering relation
		Object2ObjectOpenHashMap<Element,Set<Element>> map = new Object2ObjectOpenHashMap<Element,Set<Element>>(); 
		for( Element x : elements )
			for( Element y : elements ) {
				// Test only if x < y (strict!)
				if ( x == y || ! x.leq( y ) ) continue;
				
				boolean middle = false;
				for( Element z : elements ) {
					if ( x != z && y != z && x.leq( z ) && z.leq( y ) ) {
						middle = true;
						break;
					}
				}
				
				if ( ! middle ) {
					Set<Element> curr = map.get( x );
					if ( curr == null ) {
						curr = new ObjectOpenHashSet<Element>();
						curr.add( y );
						map.put( x, curr );
					}
					else curr.add( y );
				}
			}
		
		Assert.assertEquals( map, lattice.coveringRelation() );
	}
}
